/*

jacobi2.c

kyung-suk Lhee
11/17/98

This program is written to use jacobi iteration to solve Laplace
equations.

difference from original jacobi program
----------------------------------------

1. computing norm

  in the original program,
    diff = diff /((double) (xpoints+1) * (ypoints+1))

  in this program, diff is used as is.

2. norm is computed for every iteration, not for every 20th iteration
   Also, computing local diff is in the range of 1 ~ xpoints and
   1 ~ ypoints since these are the core data range

These changes result in smaller number of iteration.

*/

#include <mpi.h>
#include <stdio.h>
#include <math.h>


#define MAXITER 100000    /* maximum number of iterations */
#define TOL 1.0e-4        /* tolerance (error factor)     */
#define IV 0.9            /* initial value for interior of region */

#define xprocessors 2
#define yprocessors 2

#define xpoints 8
#define ypoints 8

#define NDIM 2

main(int argc, char *argv[])
{
    MPI_Status status;

    int rank, leftneighbor, rightneighbor;
    int bottomneighbor, topneighbor;
    int coords[NDIM], src;
    int rankx, ranky;


    double phi[ypoints+2][xpoints+2], oldphi[ypoints+2][xpoints+2];

    double diff=0.0,  /* the average absolute difference in elements   */
      ediff=0.0, /* the difference in two elements of phi and oldphi */
      in=0.0;    /* sum of diff from each task */
    
    int mask[ypoints+2][xpoints+2];

    int value, size, i, j, m, n, reorder;
    int dims[2], periods[2];
    
    int done = 0;     /* test flag */
    int iter = 0;     /* iteration counter */
    

    MPI_Comm comm2d;
    MPI_Datatype  contig, strided;


    MPI_Init(&argc, &argv);
    MPI_Comm_rank( MPI_COMM_WORLD, &rank );
    MPI_Comm_size( MPI_COMM_WORLD, &size );

    /* .... set up communication ....*/

    periods[0] = 0;
    periods[1] = 0;

    reorder = 0;

    dims[0] = yprocessors;
    dims[1] = xprocessors;

    MPI_Cart_create(MPI_COMM_WORLD, 2, dims, periods, reorder,
                    &comm2d);
    MPI_Cart_coords(comm2d, rank, 2, coords);

    ranky = coords[0];
    rankx = coords[1];

    MPI_Cart_shift(comm2d, 0, 1, &bottomneighbor, &topneighbor);
    MPI_Cart_shift(comm2d, 1, 1, &leftneighbor, &rightneighbor);

    /* .... set up the grid ....*/

    for (j = 0; j < xpoints+2; j++){
      for (i = 0; i < ypoints+2; i++){
        phi[i][j] = IV;
        mask[i][j] = 1;
      }
    }

    /* set mask to be false on boundary of region to be solved
     *  and phi to be the solution on the boundary.
     * For this simple test program, we don't use a real equation,
     * we just set the boundary to be 1.0, as it makes it easier
     * to observe that your iteration is working correctly
     */

    if (rankx == 0) {
      for (i = 1; i < ypoints+1; i++){ mask[i][1] = 0; 
      phi[i][1]= 1.0;
      }
    }

    if (rankx == (xprocessors-1)){  
      for (i = 1; i < ypoints+1; i++){ mask[i][xpoints] = 0;
      phi[i][xpoints]= 1.0;
      }
    }

    if (ranky == 0) {
      for (j = 1; j < xpoints+1; j++) { mask[1][j] = 0; 
        phi[1][j]= 1.0;
      }
    }

    if (ranky == (yprocessors-1)) {
      for (j = 1; j < xpoints+1; j++) { mask[ypoints][j] = 0;
      phi[ypoints][j]= 1.0;
      }
    }

    /* MPI defined types - contig for row types in C,
     *                     strided for column types in C.  */
    MPI_Type_contiguous (ypoints, MPI_DOUBLE, &contig);
    MPI_Type_vector (xpoints, 1, ypoints+2, MPI_DOUBLE, &strided);
    MPI_Type_commit (&contig);
    MPI_Type_commit (&strided);


    /* .............Iterate till done...............*/
    while (!done)
      {
        /* copy phi array to oldphi before each iteration */

        for (j = 0; j <= xpoints+1; j++)
          {
            for (i = 0; i <= ypoints+1; i++)
              {
                oldphi[i][j] = phi[i][j];
              }
          }
        /* communicate edge rows and columns to neighboring processors
         *   to put in their guard rings               */

        MPI_Sendrecv (&(oldphi[1][1]), 1, strided, leftneighbor, 30, 
                      &(oldphi[1][xpoints+1]), 1, strided, rightneighbor, 30,
                      comm2d, &status);

        MPI_Sendrecv (&(oldphi[1][xpoints]), 1, strided, rightneighbor, 31,
                      &(oldphi[1][0]), 1, strided, leftneighbor, 31, 
                      comm2d, &status);

        MPI_Sendrecv (&(oldphi[1][1]), 1, contig, topneighbor, 40, 
                      &(oldphi[ypoints+1][1]), 1, contig, 
                      bottomneighbor, 40, comm2d, &status);

        MPI_Sendrecv (&(oldphi[ypoints][1]), 1, contig, bottomneighbor, 41,
                      &(oldphi[0][1]), 1, contig, topneighbor, 41, 
                      comm2d, &status);

        /* jacobi iteration of the difference equation:
         *  for each point in phi which is in the interior of the region,
         *  defines a new value for phi in terms of old values of neighbors */

        for (j = 1; j <= xpoints; j++)
          {
            for (i = 1; i <= ypoints; i++)
              {
                if (mask[i][j]) {
                  phi[i][j] = 0.25 * (oldphi[i-1][j] + 
                                      oldphi[i+1][j] 
                                      + oldphi[i][j-1] + 
                                      oldphi[i][j+1]);
                }
              }
          }
    
        /* check for convergence every 20 iterations */
        /* find the average absolute change in elements of phi */
        /* maximum iterations is MAXITER */

        if (iter >= MAXITER) 
          done=1;
        else
          iter++;

        /*if ((fmod((double) iter, (double)20.0) == 0.0) && (!done)) */
          {
            /* compute maximum difference between elements of phi 
             *   and oldphi in each processor  */

            diff = 0.0;
            for (j = 1; j <= xpoints; j++)
              {
                for (i = 1; i <= ypoints; i++)
                  {
                    ediff = phi[i][j] - oldphi[i][j];
                    if (ediff < 0.0) 
                      ediff = -ediff;
                    if (diff < ediff) diff = ediff;
                  }
              }
            
            /*diff = diff /((double) (xpoints+1) * (ypoints+1));*/
            
            /* maximum difference over all processors */          
            MPI_Allreduce(&diff, &in, 1, MPI_DOUBLE, MPI_MAX, comm2d);

            if (in < ((double)TOL))
              done = 1;

          }
      }
    
    if (rank == 0) {
      printf ("*** Number of Iterations = %d\n", iter);
      printf ("*** max diff is = %g, tolerance is = %g\n", in, (double)TOL);
    }

    /* print results */

    for (m = 0; m <= xprocessors-1; m++){
      for (n = 0; n <= yprocessors-1; n++){
        MPI_Barrier(MPI_COMM_WORLD);

        if ((m == rankx) && (n == ranky)) 
          for (j = 1; j <= xpoints; j++){
            { 

            printf("Proc rankx=%d ranky=%d, column=%d \n", m, n, j);

            for (i = 1; i <= ypoints; i++)
              { /* print jth column from current processor */
                printf("%10.6e ", phi[i][j]);
              }
            printf("\n");
            fflush(stdout);
            }
          }
      }
    }

    MPI_Type_free(&contig);
    MPI_Type_free(&strided);

    MPI_Finalize();
}
